//
// Module: CalCoreData
// Author: Hassan Siddiqui
//
// Description: Data/methods for ARF generation specific to EPIC CAL data
//
// Comments:
//
// Modifications: See ChangeLog
//
// $Id: CalCoreData.cc,v 1.16 2014/04/03 13:37:51 rsaxton Exp $

/* General */

#include "Xmm.h"
#include "common.h"
#include "OalServer.h"
#include "CalServer.h"
#include <Region.h>
#include <ASCregion.h>
#include "Tracer.h"
#include "param.h"
#include "CcdQuantumDataServer.h"
#include "STime.h"

/* Local classes */
#include "CalCoreData.h"
#include "CoreData.h"
#include "SimpleIterator.h"
#include "NullSplitEventsData.h"
#include "NonUnifBinnedAxis.h"
#include "NullPileupCorrector.h"
#include "CirclePsfModel.h"
#include "CalPsfModel.h"

/* Constructors */

CalCoreData::CalCoreData( double psfenergy,
			  const string & datamode,
			  const string & dateobs,
			  const PosWCS * const poswcs) :
  _sourcePosition(0),
  _pfds(0),
  _psf(0),
  _fds(0),
  _cnds(0),
  _pileupCorrector(0),
  _positionUtils(0),
  _psfenergy(psfenergy),
  _datamode(datamode),
  _dateobs(dateobs),
  _poswcs(poswcs)
{
  PROC; ENDPROC;
};

/*  Destructors  */
CalCoreData::~CalCoreData()
{
  PROC;

  delete _psf; _psf = 0;
  delete _sourcePosition;  _sourcePosition = 0;
  delete _positionUtils;   _positionUtils  = 0;
  delete _pileupCorrector; _pileupCorrector = 0;

  ENDPROC;
};

/* Initialisation */
void CalCoreData::init(Xmm::Instrument instrumentID)
{
  errstr<<"CoreData init..."
                << Sas::message(Msg::App, Msg::Verbose);

  /* Create PositionUtils object */
  _positionUtils = new PositionUtils(instrumentID);

  /* Get effective area object */
  //_effds = calServer.getAtom(_effds);

  /* Get effective area at the source position - now done with the
     source position initialisation */
  //_sourceposEffectiveArea =
   // _effds->effectiveAreaCurve( _sourcePosition->theta(),
//			       _sourcePosition->phi());

  /* Get quantum efficiency object */
  _qds = calServer.getAtom(_qds);

  /* Get pattern fraction object */
  _pfds = _qds->patternFractionData();

  /* Get filter transmission object */
  _fds = calServer.getAtom(_fds);

  /* Get contaminant transmission object */
  _cnds = calServer.getAtom(_cnds);

  /* Create a PileupCorrector object */
  _pileupCorrector = new NullPileupCorrector();
}

/* Set source position */
void CalCoreData::setSourcePosition(const string & sourceCoordType,
                                 double sourcex,
                                 double sourcey)
{
  /* [HS] Check that source position has
     not been initialised previously */
  if (_sourcePosition)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }

  _sourcePosition =
    new SourcePosition(_positionUtils->
                       sourcePosition(sourceCoordType,
                                      _datamode,
                                      _dateobs,
                                      sourcex,
                                      sourcey,
                                      _poswcs));

  /* Set-up the effective area server */
  _effds = calServer.getAtom(_effds);

  /* Get the effective area curve at the source position */
  _sourceposEffectiveArea =
             _effds->effectiveAreaCurve( _sourcePosition->theta(),
	       		                   _sourcePosition->phi());

//  _sourceposQuantumEfficiency = quantumEfficiencyCurve( 
  //                       _sourcePosition->rawx(), _sourcePosition->rawy());

  //_effareaset = true;
}


void CalCoreData::setPSF( const GeoRegion &region, const RegionExtent * extent)
{
  PROC;

   errstr<<"Region in setPSF: "<<region.size()<<" "<<region[0].size()
         <<" "<<region[0][0]->type()<<" "
                <<region[region.size()-1][0]->type()
                << Sas::message(Msg::App, Msg::Verbose);

  /* Set the accuracy level for this PSF from the input region */
   AccuracyLevel accuracy = getAccuracyLevel(region);

  /* Set the PSF to use */
   _psf = new CalPsfModel( _psfenergy,
                _sourcePosition->xdet(),
                _sourcePosition->ydet(),
                accuracy,
                _positionUtils);

  /* Set the radii if using circular or elliptical mode */
   if (accuracy == EXTENDED || accuracy == ELLBETA)
   //if (accuracy == EXTENDED)
   {

  /* Find the source centre and radius in det coords and output as debug*/
      double circleX=0, circleY=0;
      circleX = (extent->xmax() + extent->xmin()) / 2.0;
      circleY = (extent->ymax() + extent->ymin()) / 2.0;
      errstr<<" Circle X: "<<circleX<<" Y: "<<circleY
            << Sas::message(Msg::App, Msg::Verbose);

      vector<double> circleRad;

  /* Get circle/elliptical/annulus radii from region */
      for (unsigned j=0; j<region[0].size(); j++) {
        for (unsigned i=0; i<region[0][j]->radii().size(); i++) {

    /* If the region is a box the radii is actually a diameter */
          if (region[0][0]->type() == ShapeBasics::BOX ||
              region[0][0]->type() == ShapeBasics::ROTBOX) {

              circleRad.push_back(
               detxy005ToCamCoord2( INT(region[0][j]->radii()[i]/2.0), 1 ).x() ); //Det - mm
         /* For ellipses we just want the biggest radius */
          } else if (region[0][0]->type() == ShapeBasics::ELLIPSE ||
                     region[0][0]->type() == ShapeBasics::ELLIPTANNULUS) {
             if (i==1 || i==3) {
                circleRad.push_back(
                 detxy005ToCamCoord2( INT(max(region[0][j]->radii()[i],
                         region[0][j]->radii()[i-1])), 1 ).x() );
             }
          } else {
              circleRad.push_back(
               detxy005ToCamCoord2( INT(region[0][j]->radii()[i]), 1 ).x() );
          }
          errstr<<" Rad: "<<region[0][j]->radii()[i] / 20.0
                << Sas::message(Msg::App, Msg::Verbose);
        }
      }

  /* Make sure that the outer radius is bigger than the inner */
      if (circleRad.size() == 2 && circleRad[0] > circleRad[1]) {
         double crtemp = circleRad[0];
         circleRad[0] = circleRad[1];
         circleRad[1] = crtemp;
      }

  /* Set the radii for this circular PSF */
      _psf->setRadius(circleRad);

  /* Define the box shape within the PSF object */
      _psf->setBoxShape(region[0][0]->type());

  /* Define the orientation within the PSF object for elliptical shapes */
      if (accuracy == ELLBETA &&
             (region[0][0]->type() == ShapeBasics::ELLIPSE ||
                     region[0][0]->type() == ShapeBasics::ELLIPTANNULUS)) {
     
         errstr<<"Setting orientation to: "<<region[0][0]->rotAngles()[0]
               << Sas::message(Msg::App, Msg::Verbose);
         _psf->setOrientation(region[0][0]->rotAngles()[0]);
      }

  // Debug
    //  errstr<<" Made Circular/Elliptical PSF "<<_sourcePosition->theta()
      //          <<" "<<_sourcePosition->phi()<<" "<<circleRad[0]
        //        <<" "<<circleRad[1]<<Sas::message(Msg::App, Msg::Verbose);
   }
}

void CalCoreData::setPSF_FMF(double factor)
{
   _psf->setFastModeFactor(factor);
}

// Set the RAW-Y value of the source centre
void CalCoreData::setSRCPOS(int SRCPOS)
{
   _SRCPOS = SRCPOS;
   _psf->setSRCPOS(SRCPOS);
}

const BinnedAxis * CalCoreData::ebins() const
{
  PROC;
  CanonicalRedist * rmf = 0; rmf = calServer.getAtom(rmf);

  RETURN new NonUnifBinnedAxis(rmf->eBins());
}

/* Return source position */
const SourcePosition * CalCoreData::sourcePosition() const
{ PROC;
  if (!_sourcePosition)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
  RETURN _sourcePosition;
}

/* Return position utility object */
const PositionUtils * CalCoreData::positionUtils() const
{ PROC;
  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }

   RETURN _positionUtils;
}
PositionUtils * CalCoreData::varpositionUtils()
{ PROC;
  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }

   RETURN _positionUtils;
}


/* Return PSF */
PsfModel * CalCoreData::psf() const
{ PROC; RETURN _psf; }

const TelCoords CalCoreData::camCoord2ToTelCoord( const CamCoords & camCoords ) const
  {
    PROC;
  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
    RETURN _positionUtils-> camCoord2ToTelCoord (camCoords);
  };

const PixCoords CalCoreData::camCoord2ToPixCoord( const CamCoords & camCoords ) const
  {
   PROC;
    if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
    RETURN  _positionUtils-> camCoord2ToPixCoord (camCoords);
  };

/* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
const CamCoords CalCoreData::detxy005ToCamCoord2( int detx,
						    int dety ) const
{
  PROC;
  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
  RETURN  _positionUtils->detxy005ToCamCoord2( double(detx),
					       double(dety) );

}

SourcePosition CalCoreData::xyToDetxy (double sourcex,
						double sourcey) const
{
  PROC;

  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }

  SourcePosition s = _positionUtils->sourcePosition( "POS",
						     _datamode,
						     _dateobs,
						     sourcex,
						     sourcey,
						     _poswcs);
  RETURN s;
}

/* Frame period, in seconds */
double CalCoreData::framePeriod()          const
{
  PROC;
  RETURN 1.0;
};


/** Iterator methods. */

GeneralIterator<int> * CalCoreData::ccdIterator()
{
  PROC;
  RETURN ( new SimpleIterator<int>( ccdFirst(),
				    ccdLast()));
}

/* Data resource methods */

/* Method: effectiveArea (double energy) 
   Purpose: Return the effective area curve at the source position 
*/
EffectiveAreaCurve CalCoreData::effectiveArea () const
{
  PROC;
  RETURN _sourceposEffectiveArea;
}

/* Method: effectiveArea (double theta, double phi, double energy) */
/* Purpose: Return the effective area curve at the input position */
EffectiveAreaCurve CalCoreData::effectiveArea (double theta, double phi) const
{
  PROC;
  RETURN _effds->effectiveAreaCurve( theta, phi);
}

/* Method: mirrorAdjustment(eacurve) */
/* Purpose: Convolve the effective area curve with the mirror adjustment*/
EffectiveAreaCurve CalCoreData::mirrorAdjustment(EffectiveAreaCurve eain) const
{
  PROC;
  RETURN _effds->applyAreaCorr(eain);
}

CcdQuantumEfficiencyCurve
CalCoreData::quantumEfficiencyCurve  ( double rawx,
				       double rawy ) const
{
  PROC;

  unsigned short rawxi = SHORT(UNT(rawx)),
                                 rawyi = SHORT(UNT(rawy));

 // If this is Timing mode override the RAW-Y position with the source
 // centre
  Xmm::Mode windowMode = calServer.state()->mode()->get();
  if (windowMode == Xmm::FAST_TIMING) rawyi = SHORT(_SRCPOS);

  CcdQuantumEfficiencyCurve qecurve = _pfds->at(rawxi, rawyi).total();

errstr<<"QETOTAL(0.3keV): "<<qecurve.efficiency(300.0)<<" raw: "<<rawxi<<" "<<rawyi
            << Sas::message(Msg::App, Msg::Verbose);

 /* Multiply by the energy pattern fraction. If grades is set to invalid then
    dont multiply by anything as this should mean that ALL patterns
    are present */

  if (_eventGrade != CombinedEventGrades::INVALID)
   {
     qecurve *= _pfds->at(rawxi, rawyi).grade(_eventGrade).inEnergySpace();
   }

errstr<<"QECURVE: "<<qecurve.efficiency(300.0)<<" raw: "<<rawxi<<" "<<rawyi
            << Sas::message(Msg::App, Msg::Verbose);

  RETURN qecurve;
}


FilterTransmissionCurve
CalCoreData::filterTransmissionCurve (double theta,
				    double phi ) const
{
  PROC;
  if (!_fds)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
  RETURN _fds->at(theta, phi);
}

ContaminantTransmissionCurve *
CalCoreData::contaminantTransmissionCurve() const
{
  PROC;
  if (!_cnds)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }

  // Find the observation time as MJD
  double time = STime(_dateobs).modifiedJulianDay().full();

  RETURN _cnds->at(time);
}

bool CalCoreData::setCcd(const CamCoords & camcoord)
{
  PROC;

  if (!_positionUtils)
    {
      errstr<<"Anomaly in CalCoreData - raise SPR"
          << Sas::error("ARFObjError");
    }
  bool result = _positionUtils->setCcd(camcoord);

  /* Even if the setCcd call resulted in no ccd found
     assume that call reset ccd number, and update state. */
  updateState();

  RETURN result;
}

const PileupCorrector * CalCoreData::pileupCorrector() const
    {
      PROC;
      if (!_pileupCorrector)
	{
	  errstr
	    << "The pile-up corrector does not appear to be present. Please raise an SPR "
	    << Sas::error("InvalidPileupCorrector");
	}

      RETURN _pileupCorrector;
    };


void CalCoreData::updateState()
    {
      PROC;

      /* Get quantum efficiency object */
        _qds = calServer.getAtom(_qds);

      /* Get pattern fraction object */
        _pfds = _qds->patternFractionData();

      ENDPROC;
    }

/* Get selected patterns from DSS */
void CalCoreData::initPatterns( DataSubspace * dss )
{
  PROC;

  /* Determine the number of patterns available */
  PatternLibServer * ptls = 0; ptls = calServer.getAtom(ptls);

  unsigned numberofpatterns = ptls->patterns();

  /* The number of patterns returned for the PN is wrong for historical
     reasons - put a fix in here so that it doesn't cause a problem */
  numberofpatterns = numberofpatterns < 32 ? 32:numberofpatterns;

  errstr
    << "Number of patterns for detector " << numberofpatterns
    << Sas::message(Msg::App, Msg::Noisy);

  bool patterntmp[numberofpatterns];
  for (unsigned j = 0; j < numberofpatterns; ++j)
    {
      patterntmp[j] = false;
    }

  double upper, lower;
  RangeBound upperB, lowerB;
  double lowest=12, highest=0;
  list <Range<double> > ranges = dss->ranges("PATTERN");

  if (ranges.size()==0)
    {
      errstr<< "No PATTERN range info found in DSS - assuming energy pattern "
            " fraction = 1.0"
            << Sas::warning("NonStandardPatterns");

      for (unsigned j = 0; j < numberofpatterns; ++j )
	{
	  _patterns.push_back(j);
	}

     /* The invalid flag is taken to mean that all patterns are present */
      _eventGrade=CombinedEventGrades::INVALID;

      ALTRETURN;
    }


  for ( list<Range<double> >::iterator i = ranges.begin();
	i != ranges.end();
	++i )
    {
      (*i).values(lower, upper, lowerB, upperB);

      if (lowerB==Unbounded)
	{
	  lower = 0;
	}
      else if (lowerB==Open)
        {
          ++lower;
        }

      if (upperB==Unbounded)
	{
	  upper  = numberofpatterns;
	}
      else if (upperB==Open)
        {
          --upper;
        }


      for ( unsigned j = (unsigned) lower;
	    j <= (unsigned) upper;
	    ++j )
	{
	  patterntmp[j]=true;
	}
    }

  for (unsigned j = 0; j < numberofpatterns; ++j )
    {
      if (patterntmp[j])
	{
	  _patterns.push_back(j);

          lowest = j < lowest ? j:lowest;
          highest = j > highest ? j:highest;
	}
    }

  errstr
    << "Number of patterns selected: " << _patterns.size()
    << Sas::message(Msg::App, Msg::Noisy);


  if (_patterns.size() == 0)
    {
        errstr<< "No patterns found in DSS - assuming energy pattern "
            " fraction = 1.0"
            << Sas::warning("NonStandardPatterns");

        _eventGrade=CombinedEventGrades::INVALID;
    }
  else {

 /* Set the event grade from the pattern range */
       if (lowest==0 && highest==0)
       {
        _eventGrade=CombinedEventGrades::SINGLE;
       }
     else if (lowest==0 && highest==4)
       {
        _eventGrade=CombinedEventGrades::SD;
       }
     else if (lowest==0 && highest==12)
       {
        _eventGrade=CombinedEventGrades::SDTQ;
       }
     else if (lowest==1 && highest==4)
       {
        _eventGrade=CombinedEventGrades::DOUBLE;
       }
     else if (lowest==5 && highest==8)
       {
        _eventGrade=CombinedEventGrades::TRIPLE;
       }
     else if (lowest==9 && highest==12)
       {
        _eventGrade=CombinedEventGrades::QUADRUPLE;
       }
     else
       {
          errstr
        << "Non-standard pattern range - assuming energy pattern "
           "fraction = 1.0"
            << Sas::warning("NonStandardPatterns");

        _eventGrade=CombinedEventGrades::INVALID;
       }

     errstr
       << "lowest: "<<lowest<<" highest: "<<highest
       << " eventGrade="<<_eventGrade
       << Sas::message(Msg::App, Msg::Verbose);
   }

  ENDPROC;
}

/* Check that the window mode is compatible with the selected patterns */
void CalCoreData::checkModePattern(Xmm::Mode windowMode)
{
  PROC;

    switch (windowMode) {

        case Xmm::FAST_TIMING:
      //  case Xmm::FAST_BURST:
          {
            if (_eventGrade != CombinedEventGrades::SD)
            {
               errstr<<"It is strongly recommended that spectra "
               <<"containing single and double pixels (patterns 0-4) "
               <<"are used when analysing PN Timing data. Other pattern "
               <<"combinations are not calibrated and will almost certainly "
               <<"produce erroneous results"
               << Sas::warning("BadModePatternCombination");
            }
          }
        default:
          {};
      }
  ENDPROC;
}

/* Check whether the region is an optimum ellipse or annular_ellipse */
bool CalCoreData::isRegionOptimumEllipse(const GeoRegion &region)
{
  PROC;

      bool optEllipse = false;

    /* Check whether the source region is an ellipse or annular ellipse */
      /* AI gcc4 porting
       *  error: suggest parentheses around && within ||
       */
      if ( ( region.size()==1 )
    		  &&
    		  ( ( ( region[0].size()==1 )
    			&&
    		       ( (region[0][0]->type() == ShapeBasics::ELLIPSE)
    				||
    			 (region[0][0]->type() == ShapeBasics::ELLIPTANNULUS)) )
    				  ||
    			   ( (region[0].size()==2)
    				        &&
    		         ( (region[0][0]->type() == ShapeBasics::ELLIPSE)
    					&&
    		           (region[0][1]->type() == ShapeBasics::ELLIPSE)
    			                &&
    		           (abs(region[0][0]->pointsX()[0] - 
                                      region[0][1]->pointsX()[0])<=1.0)
    		                        &&
    			   (abs(region[0][0]->pointsY()[0] - 
                                      region[0][1]->pointsY()[0])<=1.0)
    				        &&
    			   (region[0][0]->included() != 
                                      region[0][1]->included()) ) ) )
      )
      {

      /* Get the ellipticity of the optimum ellipse at this theta, phi
         from the CCF */
        double energy=1500.;  // Same as eregionanalyse but must set as param..
                              // in CAL !!

        calServer.state()->accuracyLevel()->set(ELLBETA);
        PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
        Psf & p = psfs->psf(energy, _sourcePosition->theta(),
                                             _sourcePosition->phi());
        double optimum_ellipticity = p.ellipticity();

      /* Get the orientation of the ellipse from the region */
        double orient = region[0][0]->rotAngles()[0];

     /* Find the optimum orientation */
        // Need a common function to do this !!!!

        double optimum_orient = orient;  // CHANGE !!!!

      /* Calculate the ellipticity of the shape */
        double ellipticity = sqrt( 1.0 -
                ( region[0][0]->radii()[1] * region[0][0]->radii()[1] ) /
                ( region[0][0]->radii()[0] * region[0][0]->radii()[0] ) );

        errstr<<"Ellipticity from shape: "<<ellipticity<<" "
              <<region[0][0]->radii()[1]<<" "<<region[0][0]->radii()[0]
              << Sas::message(Msg::App, Msg::Verbose);

      /* Check whether the ellipticity and orientation are optimum.
         Use a tolerance of 0.05 and 5 degs respectively for now */
        if ( (abs(ellipticity - optimum_ellipticity) < 0.05) &&
             (abs(orient - optimum_orient) < 5.0 ) )
           optEllipse=true;

        errstr<<"Optimum ellipse ? "<<optEllipse<<" "<<ellipticity
              <<" "<<optimum_ellipticity<<" "<<_sourcePosition->theta()
              << Sas::message(Msg::App, Msg::Verbose);
     }

     RETURN optEllipse;
}

/* Check whether a circular PSF can be used with this region */
bool CalCoreData::useCircularPSF(const GeoRegion &region)
{
  PROC;

      bool useCircular = false;

      /* AI gcc4 porting
       * error: suggest parentheses around && within ||
	   */
      if
      (
    		  region.size()==1
    		  &&
    		  (
    				  (
    						  region[0].size()==1
    						  &&
    						  (
    								  region[0][0]->type() == ShapeBasics::CIRCLE ||
    								  region[0][0]->type() == ShapeBasics::ANNULUS ||
    								  region[0][0]->type() == ShapeBasics::BOX ||
    								  region[0][0]->type() == ShapeBasics::ROTBOX
    						  )
    				  )
    				  ||
    				  (
    						  region[0].size()==2 &&
    						  region[0][0]->type() == ShapeBasics::CIRCLE &&
    						  region[0][1]->type() == ShapeBasics::CIRCLE &&
    						  abs(region[0][0]->pointsX()[0] - region[0][1]->pointsX()[0])<=1.0 &&
    						  abs(region[0][0]->pointsY()[0] - region[0][1]->pointsY()[0])<=1.0 &&
    						  region[0][0]->included() != region[0][1]->included()
    				  )
    		  )

      )
      {

          useCircular = true;
      }

      //cout<<"Use circular ?"<<useCircular<<endl;

      RETURN useCircular;
}

/* Find the appropriate PSF type from the input region */
AccuracyLevel CalCoreData::getAccuracyLevel(const GeoRegion &region)
{
  PROC;

   AccuracyLevel accuracy = HIGH;   // The default

  // Set the PSF according to the users wishes or use the best available
   if (stringParameter("psfmodel")!="notSpecified") {

     if ( uppercase(stringParameter("psfmodel")) == "HIGH") 
            accuracy = HIGH;
     else if ( uppercase(stringParameter("psfmodel")) == "MEDIUM") 
            accuracy = MEDIUM;
     else if ( uppercase(stringParameter("psfmodel")) == "EXTENDED") 
            accuracy = EXTENDED;
     else if ( uppercase(stringParameter("psfmodel")) == "ELLBETA") 
            accuracy = ELLBETA;
     // Bad input handled by configuration file

   } else { // PSF model not set on the command line

    /* If this region is a single circle or a single annulus or two circles
       with the same centre, or a box, use a circular PSF */
     if (useCircularPSF(region))
         accuracy = EXTENDED;

    /* If not then check whether this region is a valid ellipse */
     else if (isRegionOptimumEllipse(region))
         accuracy = ELLBETA;
   }

   RETURN accuracy;
}
