///
// Module: DetmapFilter
// Author: Hassan Siddiqui
//
// Description:
//
// Given an input detmap, this class generates
// a new PixelIterator for those pixels
// that are bounded by the specified region expression.
//
// Modifications: See ChangeLog
//
// $Id: DetmapFilter.cc,v 1.6 2009/09/03 13:19:51 aibarra Exp $

// General classes
#include <iostream>

// SAS-specific classes
#include "Dal.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "DataFilter.h"
#include "DataSubspace.h"

/* THIS CLASS CALLS attcalc, SO RUN AS METATASK */
#include <metatask.h>
#include "attcalc.h"

// Local classes
#include "DetmapFilter.h"
#include "TablePixelIterator.h"
#include "CompactTablePixelIterator.h"
#include "ArfDataServer.h"
#include "DatafileAttributes.h"
#include "CheckExpr.h"

/* Constructor */
DetmapFilter::DetmapFilter():
  _init(false),
  //RDS _filteredset(0),
  set(0),
  _t(0)
{
  PROC;
  ENDPROC;
}


/* Destructor */
DetmapFilter::~DetmapFilter()
{
  PROC;
  errstr
    << "Destroying DetmapFilter object .."
    << Sas::message(Msg::App, Msg::Verbose);

  if (_init)
    {
      delete _t; _t = 0;
      dataSetServer->close(_filteredset);
      //dataSetServer->close(set);
    }
  ENDPROC;
}
/* AI ggc4 porting
 * type qualifiers ignored on function return type
 * trailing const removed
 */
PixelIterator *  DetmapFilter::filteredMap( PixelIterator * const detmap,
const string & filteredtablename, DatafileAttributes * dfAttributes,
const DataSubspace * dss, bool insitu)
{
  PROC;

  if (_init) RETURN _t;

  /* Set up table name. Note that 'EVENTS' is used so that table
     can be processed by attcalc
  */
  string valuecolumnname = "VALUE";

  errstr<<" DetMapFilter called with tablename: "<<filteredtablename
   << Sas::message(Msg::App, Msg::Verbose);

  /* Create a table from the detmap, containing additional columns
   * containing the DETX DETY coords (which can be used by attcalc)
   * and the sky coordinates (generated by attcalc) for each pixel */

  // RDS
     string rawtablename    = "rawpixellist.ds:EVENTS";
  //if (! insitu) {

     set = dataSetServer->open(rawtablename, DataSet::Create);
  //} else {
  //rawtablename = filteredtablename;
  //if (dfAttributes->keepFilteredTable()) {
   //  set = dataSetServer->open(filteredtablename, DataSet::Create);
  //}
  //else {
   //  set = dataSetServer->open(filteredtablename, DataSet::Temp);
  //}
  //}

  Table * table = set->addTable(rawtablename, detmap->size());
  // end - RDS

  errstr << "Table size is: "<<detmap->size()
         << Sas::message(Msg::App, Msg::Verbose);
  Column * valueCol  = table->addColumn(valuecolumnname,
					Column::Real64,
					"Value",
					"");

  Column * skyxCol    = table->addColumn("X",
					 Column::Int32,
					 "Sky X",
					 "");


  Column * skyyCol    = table->addColumn("Y",
					 Column::Int32,
					 "Sky Y",
					 "");

  Column * detxCol    = table->addColumn("DETX",
					 Column::Int16,
					 "x det coords",
					 "0.05 arcsec");


  Column * detyCol    = table->addColumn("DETY",
					 Column::Int16,
					 "y det coords",
					 "0.05 arcsec");

  Column * indexxCol    = table->addColumn("XPIXEL",
					 Column::Int16,
					 "x index",
					 "");

  Column * indexyCol    = table->addColumn("YPIXEL",
					 Column::Int16,
					 "y index",
					 "");

  /* Have to add following entities for attcalc to work */
  Column * timeCol    = table->addColumn("TIME",
					 Column::Real64,
					 "elasped time",
					 "sec");

  * set->addAttribute("TELESCOP",    "mission name") =
    dfAttributes->telescope();

  * set->addAttribute("INSTRUME",    "instrument name") =
    dfAttributes->instrument();

  * set->addAttribute("DATAMODE",    "instrument mode") =
    dfAttributes->datamode();

  * set->addAttribute("DATE-OBS",    "Observation start date") =
    dfAttributes->observationDate();

  * set->addAttribute("DATE-END",    "Observation end date") =
    dfAttributes->observationDateEnd();

  * set->addAttribute("FILTER",      "Filter ID") =
    dfAttributes->filter();

  /* Get the windowing mode */

  * set->addAttribute("SUBMODE",      "Observing mode") =
    dfAttributes->observingMode();

  * set->addAttribute("EMDHLOW",      "EMDHLOW") = dfAttributes->emdhlow();
  * set->addAttribute("EMDHUPP",      "EMDHUPP") = dfAttributes->emdhupp();

  /* The attcalc task needs the following to be set */
  * set->addAttribute("CCDID",        "") = 1;
  * set->addAttribute("CCDNODE",        "") = 0;
  * set->addAttribute("QUADRANT",        "") = 0;

  ColumnData * timeColData   = timeCol->data();
  real64     * timeData      = timeColData->real64Data();

  ColumnData * valueColData  = valueCol->data();
  real64     * valueData     = valueColData->real64Data();

  ColumnData * skyxColData   = skyxCol->data();
  int32      * skyxData      = skyxColData->int32Data();

  ColumnData * skyyColData   = skyyCol->data();
  int32      * skyyData      = skyyColData->int32Data();

  ColumnData * detxColData   = detxCol->data();
  int16      * detxData      = detxColData->int16Data();

  ColumnData * detyColData   = detyCol->data();
  int16      * detyData      = detyColData->int16Data();

  ColumnData * indexxColData   = indexxCol->data();
  int16      * indexxData      = indexxColData->int16Data();

  ColumnData * indexyColData   = indexyCol->data();
  int16      * indexyData      = indexyColData->int16Data();

  /* This conversion is HARDWIRED, but is probably acceptable
     as the DET coord definition is concrete. */
  double det2deg=1./(72000);

  /* Add WCS attributes to the DET columns (for attcalc) */
  WCSConvertor("DETX", 0, 0, det2deg, "DETY", 0, 0, det2deg).
                             addAttributesTo(detxCol, detyCol);

  /* Add WCS attributes to the index columns
     (for pixel-DET coord mapping; this way the WCS info
     can also store the pixel size) */
  detmap->wcsconvertor()->addAttributesTo(indexxCol, indexyCol);

  /* Compute const time value to write to each row */
  double time = STime(dfAttributes->observationDate())-Xmm::missionReferenceTime;

  /* Now iterate through detmap,
     adding an entry for each and every pixel */
  int row = 0;

  FOREACH(detmap)
    {
      PixelElement pixel = detmap->currentItem();
      timeData[row] = time;
      valueData[row]  = pixel.value();
      skyxData[row]   = 0;
      skyyData[row]   = 0;
      detxData[row]   = (int) pixel.xCentre();
      detyData[row]   = (int) pixel.yCentre();
      indexxData[row] = pixel.xIndex();
      indexyData[row] = pixel.yIndex();
      ++row;
    }

  /* Close dataset  */
  delete indexyColData;
  delete indexxColData;
  delete detyColData;
  delete detxColData;
  delete skyyColData;
  delete skyxColData;
  delete valueColData;
  delete timeColData;
  dataSetServer->close(set);

  /* Run attcalc if the DSS contains components
     defined in terms of the POS axes or in the AXIS2D1 space */
  //const DataSubspace * dss = arfDataServer->dataSubspace();

  DSS::DSaxes axestest;
  axestest.push_back("POS");
  axestest.push_back("AXIS2D1");
  axestest.push_back("AXIS2D2");
  axestest.push_back("AXIS2D3");
  if (!CheckExpr::check(dss->selectExpression(axestest)))
      {

  /* Call ATTCALC to populate sky coordinate values */

  errstr << " Running attcalc on " << filteredtablename
	 << Sas::message(Msg::App, Msg::Verbose);

  /* ATTCALC changes the CAL accuracy mode (and who knows what else)
     so must save and restore this afterwards */
  AccuracyLevel tempmode = calServer.state()->accuracyLevel()->get();

  /* Open the attcalc task, set the parameters and execute the task */
  Info::Task* attcalc = task("attcalc");
  ParameterInfo& attcalcpars = *attcalc->parameters();
  attcalcpars.revertToDefault();

  attcalcpars["eventset"]->setValue(rawtablename);
  attcalcpars["attitudelabel"]->setValue("fixed");
  attcalcpars["fixedra"]->setValue(dfAttributes->rapnt());
  attcalcpars["fixeddec"]->setValue(dfAttributes->decpnt());
  attcalcpars["fixedposangle"]->setValue(dfAttributes->papnt());
  attcalcpars["refpointlabel"]->setValue("user");
  attcalcpars["nominalra"]->setValue(arfDataServer->raref());
  attcalcpars["nominaldec"]->setValue(arfDataServer->decref());
  attcalcpars["withatthkset"]->setValue(false);
  attcalc->exec();

  /* Restore the CAL accuracy mode */
  calServer.state()->accuracyLevel()->set(tempmode);

      }
  else
    {
      errstr
	<< "There does not seem to be any POS-related filters in the dss ["<<dss->selectExpression()
	<<". Will not run attcalc."
	<< Sas::message(Msg::App, Msg::Verbose);

    }

  /* If the insitu flag is set filter the exisiting table - if not
     use the command line parameters to create a new file */
       if (arfDataServer->keepFilteredTable())
	{
	  _filteredset = dataSetServer->clone( rawtablename,
				      filteredtablename,
				      Dal::Modify);
	}
       else
	{
	  _filteredset = dataSetServer->clone( rawtablename,
				      filteredtablename,
				      Dal::Temp);

   /* Remove the raw table file */
          dataSetServer->remove(rawtablename);
	}
    //}

   /* Now filter the dataset if required: filtering is Destructive */
    if (arfDataServer->filterDSS())
	{

	  /* Project region onto X,Y, DETX, DETY space */
	  DSS::DSaxes axes;
	  axes.push_back("POS");
	  axes.push_back("DET");
	  //axes.push_back("AXIS2D");
	  axes.push_back("AXIS2D1");
	  axes.push_back("AXIS2D2");
	  axes.push_back("AXIS2D3");
	  string projregionexpr = dss->selectExpression(axes);

	  errstr << "DetmapFilter: SLCTEXPR gives \'"
		 << dfAttributes->selectionExpression() << "\'"
		 << Sas::message(Msg::App, Msg::Verbose);

	  errstr << "DetmapFilter:  Filtering " << filteredtablename
		 << " using expression \'" << projregionexpr << "\'"
		 << Sas::message(Msg::App, Msg::Verbose);

	  DataFilter filter(projregionexpr);

         //RDS
	  table = _filteredset->table("EVENTS");
	  //table = set->table("EVENTS");

	  long rows = filter.doFilter(*table);

	  errstr << "DetmapFilter: got " << rows << " in filtered table."
		 << Sas::message(Msg::App, Msg::Verbose);

	}
      else
	{
	  errstr
	    << "No filtering of pixel list has been performed."
	    << Sas::message(Msg::App, Msg::Verbose);
	}

      /* Pass filterSet to pixel iterator */
      if (insitu) {

       /* For the raw pixel iterator, everything is defined in INT16.
          Use the DETX column as a dummy for the valuecolumn */

         string nullcol;
         //RDS _t = new CompactTablePixelIterator( filteredtablename+":EVENTS",
         _t = new CompactTablePixelIterator( filteredtablename,
                  "XPIXEL", "YPIXEL", "DETX", nullcol);
      } else {
         //RDS
         _t = new TablePixelIterator( filteredtablename,
				"XPIXEL",
				"YPIXEL",
				valuecolumnname
				);
         _t->normalise();
      }

      _init=true;

      RETURN _t;
}


